from xml.sax.saxutils import escape
from .utils import *
from .enums import Alignment
import warnings
import typing

class TextSetting:
    """Formatting for slices of a :class:`manim.mobject.svg.text_mobject.Text` object."""
    def __init__(
        self,
        start:int,
        end:int,
        font:str,
        slant,
        weight,
        line_num=-1,
        color: str=None,
    ):
        self.start = start
        self.end = end
        self.font = font
        self.slant = slant
        self.weight = weight
        self.line_num = line_num
        self.color = color


def text2svg(
    settings:list,
    size:int,
    line_spacing:int,
    disable_liga:bool,
    file_name:str,
    START_X:int,
    START_Y:int,
    width:int,
    height:int,
    orig_text:str,
    pango_width: typing.Union[int, None] = None,
) -> int:
    """Render an SVG file from a :class:`manim.mobject.svg.text_mobject.Text` object."""
    cdef cairo_surface_t* surface
    cdef cairo_t* cr
    cdef PangoFontDescription* font_desc
    cdef PangoLayout* layout
    cdef double font_size_c = size
    cdef cairo_status_t status
    cdef int temp_width

    file_name_bytes = file_name.encode("utf-8")
    surface = cairo_svg_surface_create(file_name_bytes,width,height)

    if surface == NULL:
        raise MemoryError("Cairo.SVGSurface can't be created.")

    cr = cairo_create(surface)
    status = cairo_status(cr)

    if cr == NULL or status == CAIRO_STATUS_NO_MEMORY:
        cairo_destroy(cr)
        cairo_surface_destroy(surface)
        raise MemoryError("Cairo.Context can't be created.")
    elif status != CAIRO_STATUS_SUCCESS:
        cairo_destroy(cr)
        cairo_surface_destroy(surface)
        raise Exception(cairo_status_to_string(status))

    cairo_move_to(cr,START_X,START_Y)
    offset_x = 0
    last_line_num = 0

    layout = pango_cairo_create_layout(cr)

    if layout == NULL:
        cairo_destroy(cr)
        cairo_surface_destroy(surface)
        raise MemoryError("Pango.Layout can't be created from Cairo Context.")

    if pango_width is None:
        pango_layout_set_width(layout, pango_units_from_double(width))
    else:
        pango_layout_set_width(layout, pango_units_from_double(pango_width))

    for setting in settings:
        family = setting.font.encode('utf-8')
        style = PangoUtils.str2style(setting.slant)
        weight = PangoUtils.str2weight(setting.weight)
        color = setting.color
        text_str = orig_text[setting.start : setting.end].replace("\n", " ")
        font_desc = pango_font_description_new()
        if font_desc==NULL:
            cairo_destroy(cr)
            cairo_surface_destroy(surface)
            g_object_unref(layout)
            raise MemoryError("Pango.FontDesc can't be created.")
        pango_font_description_set_size(font_desc, pango_units_from_double(font_size_c))
        if family:
            pango_font_description_set_family(font_desc, family)
        pango_font_description_set_style(font_desc, style.value)
        pango_font_description_set_weight(font_desc, weight.value)
        pango_layout_set_font_description(layout, font_desc)
        pango_font_description_free(font_desc)
        if setting.line_num != last_line_num:
            offset_x = 0
            last_line_num = setting.line_num
        cairo_move_to(cr,START_X + offset_x,START_Y + line_spacing * setting.line_num)

        pango_cairo_update_layout(cr,layout)
        markup = escape(text_str)
        if color:
            markup = (f"<span color='{color}'>{markup}</span>")
            if MarkupUtils.validate(markup):
                cairo_destroy(cr)
                cairo_surface_destroy(surface)
                g_object_unref(layout)
                raise ValueError(f"Pango cannot recognize your color '{color}' for text '{text_str}'.")
        if disable_liga:
            markup = f"<span font_features='liga=0,dlig=0,clig=0,hlig=0'>{markup}</span>"
        pango_layout_set_markup(layout, markup.encode('utf-8'), -1)
        pango_cairo_show_layout(cr, layout)
        pango_layout_get_size(layout,&temp_width,NULL)
        offset_x += pango_units_to_double(temp_width)

    status = cairo_status(cr)

    if cr == NULL or status == CAIRO_STATUS_NO_MEMORY:
        cairo_destroy(cr)
        cairo_surface_destroy(surface)
        g_object_unref(layout)
        raise MemoryError("Cairo.Context can't be created.")
    elif status != CAIRO_STATUS_SUCCESS:
        cairo_destroy(cr)
        cairo_surface_destroy(surface)
        g_object_unref(layout)
        raise Exception(cairo_status_to_string(status).decode())

    cairo_destroy(cr)
    cairo_surface_destroy(surface)
    g_object_unref(layout)
    return file_name

class MarkupUtils:
    @staticmethod
    def validate(markup: str) -> str:
        """Validates whether markup is a valid Markup
        and return the error's if any.

        Parameters
        ==========
        markup : :class:`str`
            The markup which should be checked.

        Returns
        =======
        :class:`str`
            Returns empty string if markup is valid. If markup
            contains error it return the error message.

        """
        cdef GError *err = NULL
        text_bytes = markup.encode("utf-8")
        res = pango_parse_markup(
            text_bytes,
            -1,
            0,
            NULL,
            NULL,
            NULL,
            &err
        )
        if res:
            return ""
        else:
            message = <bytes>err.message
            g_error_free(err)
            return message.decode('utf-8')

    @staticmethod
    def text2svg(
        text: str,
        font: str,
        slant: str,
        weight: str,
        size: int,
        _: int, # for some there was a keyword here.
        disable_liga: bool,
        file_name: str,
        START_X: int,
        START_Y: int,
        width: int,
        height: int,
        *, # keyword only arguments below
        justify: bool = None,
        indent: float = None,
        line_spacing: float = None,
        alignment: Alignment = None,
        pango_width: typing.Union[int, None] = None,
    ) -> int:
        """Render an SVG file from a :class:`manim.mobject.svg.text_mobject.MarkupText` object."""
        cdef cairo_surface_t* surface
        cdef cairo_t* context
        cdef PangoFontDescription* font_desc
        cdef PangoLayout* layout
        cdef cairo_status_t status
        cdef double font_size = size
        cdef int temp_int # a temporary C integer for conversion

        file_name_bytes = file_name.encode("utf-8")

        if disable_liga:
            text_bytes = f"<span font_features='liga=0,dlig=0,clig=0,hlig=0'>{text}</span>".encode("utf-8")
        else:
            text_bytes = text.encode("utf-8")

        surface = cairo_svg_surface_create(file_name_bytes,width,height)
        if surface == NULL:
            raise MemoryError("Cairo.SVGSurface can't be created.")
        context = cairo_create(surface)
        status = cairo_status(context)
        if context == NULL or status == CAIRO_STATUS_NO_MEMORY:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            raise MemoryError("Cairo.Context can't be created.")
        elif status != CAIRO_STATUS_SUCCESS:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            raise Exception(cairo_status_to_string(status))

        cairo_move_to(context,START_X,START_Y)
        layout = pango_cairo_create_layout(context)
        if layout == NULL:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            raise MemoryError("Pango.Layout can't be created from Cairo Context.")

        if pango_width is None:
            pango_layout_set_width(layout, pango_units_from_double(width))
        else:
            pango_layout_set_width(layout, pango_units_from_double(pango_width))

        if justify:
            pango_layout_set_justify(layout, justify)

        if indent:
            temp_int = pango_units_from_double(indent)
            pango_layout_set_indent(layout, temp_int)

        if line_spacing:
            # Typical values are: 0, 1, 1.5, 2.
            ret = set_line_width(layout, line_spacing)
            if not ret:
                # warn that line spacing don't work
                # because of old Pango version they
                # have
                warnings.warn(
                    "Pango Version<1.44 found."
                    "Impossible to set line_spacing."
                    "Expect Ugly Output."
                )

        if alignment:
            pango_layout_set_alignment(layout, alignment.value)

        font_desc = pango_font_description_new()
        if font_desc==NULL:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            g_object_unref(layout)
            raise MemoryError("Pango.FontDesc can't be created.")
        pango_font_description_set_size(font_desc, pango_units_from_double(font_size))
        if font is not None and len(font) != 0:
            pango_font_description_set_family(font_desc, font.encode("utf-8"))
        pango_font_description_set_style(font_desc, PangoUtils.str2style(slant).value)
        pango_font_description_set_weight(font_desc, PangoUtils.str2weight(weight).value)
        pango_layout_set_font_description(layout, font_desc)
        pango_font_description_free(font_desc)

        cairo_move_to(context,START_X,START_Y)
        pango_cairo_update_layout(context,layout)
        pango_layout_set_markup(layout,text_bytes,-1)
        pango_cairo_show_layout(context, layout)

        status = cairo_status(context)
        if context == NULL or status == CAIRO_STATUS_NO_MEMORY:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            g_object_unref(layout)
            raise MemoryError("Cairo.Context can't be created.")
        elif status != CAIRO_STATUS_SUCCESS:
            cairo_destroy(context)
            cairo_surface_destroy(surface)
            g_object_unref(layout)
            raise Exception(cairo_status_to_string(status).decode())

        cairo_destroy(context)
        cairo_surface_destroy(surface)
        g_object_unref(layout)
        return file_name

cpdef str pango_version():
    return pango_version_string().decode('utf-8')

cpdef str cairo_version():
    return cairo_version_string().decode('utf-8')



# -*- coding: utf-8 -*-
import os
import platform
import shutil
from pathlib import Path


def set_dll_search_path():
    # Python 3.8 no longer searches for DLLs in PATH, so we have to add
    # everything in PATH manually.
    if (
        os.name != "nt"
        or "GCC" in platform.python_compiler()
        or not hasattr(os, "add_dll_directory")
    ):
        return
    for p in os.environ.get("PATH", "").split(os.pathsep):
        try:
            os.add_dll_directory(p)
        except OSError:
            pass


def delete_media_dir():
    a = Path("media")  # absolute to the running dir
    if a.exists():
        shutil.rmtree(a)


set_dll_search_path()
delete_media_dir()
CASES_DIR = Path(Path(__file__).parent, "cases").absolute()
FONT_DIR = Path(__file__).parent / "fonts"



# -*- coding: utf-8 -*-
import os
import sys

from ._version import __version__  # noqa: F403,F401

if os.name == "nt":  # pragma: no cover
    os.environ["PATH"] = (
        f"{os.path.abspath(os.path.dirname(__file__))}"
        f"{os.pathsep}"
        f"{os.environ['PATH']}"
    )
try:
    from .cmanimpango import *  # noqa: F403,F401
    from .enums import *  # noqa: F403,F401
    from .register_font import *  # noqa: F403,F401
except ImportError as ie:  # pragma: no cover
    py_ver = ".".join(map(str, sys.version_info[:3]))
    msg = f"""

ManimPango could not import and load the necessary shared libraries.
This error may occur when ManimPango and its dependencies are improperly set up.
Please make sure the following versions are what you expect:

    * ManimPango v{__version__}, Python v{py_ver}

If you believe there is a greater problem,
feel free to contact us or create an issue on GitHub:

    * Discord: https://discord.gg/mMRrZQW
    * GitHub: https://github.com/ManimCommunity/ManimPango/issues

Original error: {ie}
"""
    raise ImportError(msg)



# -*- coding: utf-8 -*-
import sys
from textwrap import dedent

import pytest

import manimpango


def pytest_report_header(config):
    return (
        f"ManimPango version {manimpango.__version__}\n"
        f"Pango version {manimpango.pango_version()}\n"
        f"Cairo version {manimpango.cairo_version()}"
    )


@pytest.fixture
def setup_fontconfig(tmpdir, monkeypatch):
    if sys.platform.startswith("win32") and "GCC" not in sys.version:
        default_fontconfig_configuration = dedent(
            """\
        <?xml version="1.0"?>
        <!DOCTYPE fontconfig SYSTEM "urn:fontconfig:fonts.dtd">
        <!-- /etc/fonts/fonts.conf file to configure system font access -->
        <fontconfig>
            <description>Default configuration file</description>

        <!--
            DO NOT EDIT THIS FILE.
            IT WILL BE REPLACED WHEN FONTCONFIG IS UPDATED.
            LOCAL CHANGES BELONG IN 'local.conf'.

            The intent of this standard configuration file is to be adequate for
            most environments.  If you have a reasonably normal environment and
            have found problems with this configuration, they are probably
            things that others will also want fixed.  Please submit any
            problems to the fontconfig bugzilla system located at fontconfig.org

            Note that the normal 'make install' procedure for fontconfig is to
            replace any existing fonts.conf file with the new version.  Place
            any local customizations in local.conf which this file references.

            Keith Packard
        -->

        <!-- Font directory list -->

            <dir>WINDOWSFONTDIR</dir>

            <dir prefix="xdg">fonts</dir>
            <!-- the following element will be removed in the future -->
            <dir>~/.fonts</dir>

        <!--
        Accept deprecated 'mono' alias, replacing it with 'monospace'
        -->
            <match target="pattern">
                <test qual="any" name="family">
                    <string>mono</string>
                </test>
                <edit name="family" mode="assign" binding="same">
                    <string>monospace</string>
                </edit>
            </match>

        <!--
        Accept alternate 'sans serif' spelling, replacing it with 'sans-serif'
        -->
            <match target="pattern">
                <test qual="any" name="family">
                    <string>sans serif</string>
                </test>
                <edit name="family" mode="assign" binding="same">
                    <string>sans-serif</string>
                </edit>
            </match>

        <!--
        Accept deprecated 'sans' alias, replacing it with 'sans-serif'
        -->
            <match target="pattern">
                <test qual="any" name="family">
                    <string>sans</string>
                </test>
                <edit name="family" mode="assign" binding="same">
                    <string>sans-serif</string>
                </edit>
            </match>
        <!--
        Accept alternate 'system ui' spelling, replacing it with 'system-ui'
        -->
            <match target="pattern">
                <test qual="any" name="family">
                    <string>system ui</string>
                </test>
                <edit name="family" mode="assign" binding="same">
                    <string>system-ui</string>
                </edit>
            </match>

        <!--
        Load local system customization file
        -->
            <include ignore_missing="yes">conf.d</include>

        <!-- Font cache directory list -->

            <cachedir>LOCAL_APPDATA_FONTCONFIG_CACHE</cachedir>
            <cachedir prefix="xdg">fontconfig</cachedir>
            <!-- the following element will be removed in the future -->
            <cachedir>~/.fontconfig</cachedir>

            <config>
        <!--
        Rescan configuration every 30 seconds when FcFontSetList is called
        -->
                <rescan>
                    <int>30</int>
                </rescan>
            </config>

        </fontconfig>
        """
        )
        with open(tmpdir / "fonts.conf", "w") as f:
            f.write(default_fontconfig_configuration)
        monkeypatch.setenv("FONTCONFIG_PATH", str(tmpdir))
    monkeypatch.setenv("PANGOCAIRO_BACKEND", "fc")



from enum import Enum
from pango cimport *
class Style(Enum):
    """
    An enumeration specifying the various slant styles possible for a font.

    Attributes
    ----------

    NORMAL :
        the font is upright.

    ITALIC :
        the font is slanted, but in a roman style.

    OBLIQUE:
        the font is slanted in an italic style.
    """
    NORMAL = PANGO_STYLE_NORMAL
    ITALIC = PANGO_STYLE_ITALIC
    OBLIQUE = PANGO_STYLE_OBLIQUE

class Weight(Enum):
    """
    An enumeration specifying the weight (boldness) of a font.
    This is a numerical value ranging from 100 to 1000, but there are some predefined values
    Using numerical value other then that defined here is not supported.

    Attributes
    ----------

    NORMAL :
        the default weight (= 400)

    BOLD :
        the bold weight( = 700)

    THIN :
        the thin weight( = 100; Since: 1.24)

    ULTRALIGHT :
        the ultralight weight( = 200)

    LIGHT :
        the light weight( = 300)

    BOOK :
        the book weight( = 380; Since: 1.24)

    MEDIUM :
        the normal weight( = 500; Since: 1.24)

    SEMIBOLD :
        the semibold weight( = 600)

    ULTRABOLD :
        the ultrabold weight( = 800)

    HEAVY :
        the heavy weight( = 900)

    ULTRAHEAVY :
        the ultraheavy weight( = 1000; Since: 1.24)
    """
    NORMAL = PANGO_WEIGHT_NORMAL
    BOLD = PANGO_WEIGHT_BOLD
    THIN = PANGO_WEIGHT_THIN
    ULTRALIGHT = PANGO_WEIGHT_ULTRALIGHT
    LIGHT = PANGO_WEIGHT_LIGHT
    BOOK = PANGO_WEIGHT_BOOK
    MEDIUM = PANGO_WEIGHT_MEDIUM
    SEMIBOLD = PANGO_WEIGHT_SEMIBOLD
    ULTRABOLD = PANGO_WEIGHT_ULTRABOLD
    HEAVY = PANGO_WEIGHT_HEAVY
    ULTRAHEAVY = PANGO_WEIGHT_ULTRAHEAVY

class Variant(Enum):
    """
    An enumeration specifying capitalization variant of the font.

    Attributes
    ----------

    NORMAL :
        A normal font.

    SMALL_CAPS :
        A font with the lower case characters replaced by smaller variants
        of the capital characters.
    """
    NORMAL = PANGO_VARIANT_NORMAL
    SMALL_CAPS = PANGO_VARIANT_SMALL_CAPS

class Alignment(Enum):
    """
    An enumeration specifying alignment.

    Attributes
    ----------

    NORMAL :
        A normal font.

    SMALL_CAPS :
        A font with the lower case characters replaced by smaller variants
        of the capital characters.
    """
    LEFT = PANGO_ALIGN_LEFT
    CENTER = PANGO_ALIGN_CENTER
    RIGHT = PANGO_ALIGN_RIGHT



# -*- coding: utf-8 -*-
import pytest

from manimpango.enums import Style, Weight
from manimpango.utils import PangoUtils


def test_str2style():
    p = PangoUtils()

    assert p.str2style("NORMAL") == Style.NORMAL
    assert p.str2style("ITALIC") == Style.ITALIC
    assert p.str2style("OBLIQUE") == Style.OBLIQUE

    with pytest.raises(AttributeError) as exc_info:
        p.str2style("DNE")
    exception_raised = exc_info.value
    assert "There is no Style Called DNE" == str(exception_raised)


def test_str2weight():
    p = PangoUtils()

    assert p.str2weight("NORMAL") == Weight.NORMAL
    assert p.str2weight("BOLD") == Weight.BOLD
    assert p.str2weight("THIN") == Weight.THIN
    assert p.str2weight("THIN") == Weight.THIN
    assert p.str2weight("ULTRALIGHT") == Weight.ULTRALIGHT
    assert p.str2weight("LIGHT") == Weight.LIGHT
    assert p.str2weight("BOOK") == Weight.BOOK
    assert p.str2weight("MEDIUM") == Weight.MEDIUM
    assert p.str2weight("SEMIBOLD") == Weight.SEMIBOLD
    assert p.str2weight("ULTRABOLD") == Weight.ULTRABOLD
    assert p.str2weight("HEAVY") == Weight.HEAVY
    assert p.str2weight("ULTRAHEAVY") == Weight.ULTRAHEAVY

    with pytest.raises(AttributeError) as exc_info:
        p.str2weight("DNE")
    exception_raised = exc_info.value
    assert "There is no Font Weight Called DNE" == str(exception_raised)



from glib cimport *
from cairo cimport *
cdef extern from "pango/pangocairo.h":
    int PANGO_SCALE
    int pango_units_from_double(double d)
    double pango_units_to_double (int i)
    ctypedef struct PangoLayout:
        pass
    ctypedef struct PangoContext:
        pass
    ctypedef struct PangoFontMap:
        pass
    ctypedef struct PangoFontDescription:
        pass
    ctypedef struct PangoFontFamily:
        pass
    ctypedef enum PangoStyle:
        PANGO_STYLE_NORMAL
        PANGO_STYLE_OBLIQUE
        PANGO_STYLE_ITALIC
    ctypedef enum PangoWeight:
        PANGO_WEIGHT_THIN
        PANGO_WEIGHT_ULTRALIGHT
        PANGO_WEIGHT_LIGHT
        PANGO_WEIGHT_BOOK
        PANGO_WEIGHT_NORMAL
        PANGO_WEIGHT_MEDIUM
        PANGO_WEIGHT_SEMIBOLD
        PANGO_WEIGHT_BOLD
        PANGO_WEIGHT_ULTRABOLD
        PANGO_WEIGHT_HEAVY
        PANGO_WEIGHT_ULTRAHEAVY
    ctypedef enum PangoVariant:
        PANGO_VARIANT_NORMAL
        PANGO_VARIANT_SMALL_CAPS
    ctypedef enum PangoWrapMode:
        PANGO_WRAP_WORD
        PANGO_WRAP_CHAR
        PANGO_WRAP_WORD_CHAR
    ctypedef enum PangoAlignment:
        PANGO_ALIGN_LEFT
        PANGO_ALIGN_CENTER
        PANGO_ALIGN_RIGHT
    PangoLayout* pango_cairo_create_layout(cairo_t* cr)
    void pango_cairo_show_layout(
        cairo_t* cr,
        PangoLayout* layout
    )
    void pango_cairo_update_layout(
        cairo_t* cr,
        PangoLayout* layout
    )
    PangoFontDescription* pango_font_description_new()
    void pango_font_description_set_size(
        PangoFontDescription* desc,
        gint size
    )
    void pango_font_description_set_family(
        PangoFontDescription* desc,
        const char* family
    )
    void pango_font_description_set_style(
        PangoFontDescription* desc,
        PangoStyle style
    )
    void pango_font_description_set_weight(
        PangoFontDescription* desc,
        PangoWeight weight
    )
    void pango_font_description_set_variant(
        PangoFontDescription* desc,
        PangoVariant variant
    )

    void pango_layout_set_width(
        PangoLayout* layout,
        int width
    )
    void pango_layout_set_font_description(
        PangoLayout* layout,
        const PangoFontDescription* desc
    )
    void pango_layout_set_text(
        PangoLayout* layout,
        const char* text,
        int length
    )
    void pango_layout_set_wrap(
        PangoLayout *layout,
        PangoWrapMode wrap
    )
    void pango_layout_set_markup(
        PangoLayout *layout,
        const char *markup,
        int length
    )
    void pango_layout_get_size(
        PangoLayout* layout,
        int* width,
        int* height
    )
    const char* pango_version_string()
    void pango_font_description_free(
        PangoFontDescription *desc
    )
    gboolean pango_parse_markup(
       const char *markup_text,
       int length,
       unsigned int accel_marker,
       void* attr_list,
       void* text,
       void* accel_char,
       void* error
    )
    PangoFontMap* pango_cairo_font_map_new()
    void pango_font_map_list_families(
        PangoFontMap *fontmap,
        PangoFontFamily ***families,
        int *n_families
    )
    PangoFontFamily* pango_font_map_get_family(
        PangoFontMap *fontmap,
        const char *name
    )
    const char* pango_font_family_get_name(
        PangoFontFamily *family
    )
    void pango_layout_set_justify(
        PangoLayout *layout,
        gboolean justify
    )
    void pango_layout_set_indent(
        PangoLayout *layout,
        int indent
    )
    void pango_layout_set_alignment(
        PangoLayout *layout,
        PangoAlignment alignment
    )

cdef extern from *:
    """
    #if PANGO_VERSION_CHECK(1,44,0)
        int set_line_width(PangoLayout *layout,float spacing)
        {
          pango_layout_set_line_spacing(layout, spacing);
          return 1;
        }
    #else
        int set_line_width(PangoLayout *layout,float spacing){return 0;}
    #endif
    """
    # The above docs string is C which is used to
    # check for the Pango Version there at run time.
    # pango_layout_set_line_spacing is only avaiable only for
    # pango>=1.44.0 but we support pango>=1.30.0 that why this
    # conditionals.
    bint set_line_width(PangoLayout *layout,float spacing)



# -*- coding: utf-8 -*-
from pathlib import Path

import pytest

from . import CASES_DIR
from ._manim import Text
from .svg_tester import SVGStyleTester


def test_invalid_color_fail(tmpdir):
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    with pytest.raises(ValueError):
        Text("color", color="invalid", filename=str(loc))


def test_colors(tmpdir):
    expected = Path(CASES_DIR, "color_red.svg")
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    Text("color", color="red", filename=str(loc))
    assert loc.exists()
    s = SVGStyleTester(gotSVG=loc, expectedSVG=expected)
    assert len(s.got_svg_style) == len(s.expected_svg_style)
    assert s.got_svg_style == s.expected_svg_style



# -*- coding: utf-8 -*-
from pathlib import Path
from typing import Dict
from xml.parsers.expat import ParserCreate


class SVGUtils:
    SVG_DEFAULT_ATTRIBUTES: Dict[str, str] = {
        "fill": "black",
        "fill-opacity": "1",
        "stroke": "none",
        "stroke-opacity": "1",
    }

    def __init__(self) -> None:
        pass

    def parse_color_string(self, color_spec: str) -> str:
        # copied from Manim's SVG parser.
        if color_spec[0:3] == "rgb":
            # these are only in integer form,
            # but the Colour module wants them in floats.
            splits = color_spec[4:-1].split(",")
            if splits[0][-1] == "%":
                # if the last character of the first number is a percentage,
                # then interpret the number as a percentage
                parsed_rgbs = [float(i[:-1]) / 100.0 for i in splits]
            else:
                parsed_rgbs = [int(i) / 255.0 for i in splits]

            # hex_color = rgb_to_hex(parsed_rgbs)

        else:
            # not necessary for web2hex case and hex
            raise NotImplementedError

        return parsed_rgbs

    def str2styledict(self, style_specs: str) -> Dict[str, str]:
        # helper for parse style
        # copied from manim's svg parser.
        style = {}

        # the style attribute should be handled separately in order to
        # break it up nicely. furthermore, style takes priority over other
        # attributes in the same element.
        if style_specs:
            for style_spec in style_specs.split(";"):
                try:
                    key, value = style_spec.split(":")
                except ValueError as e:
                    if not style_spec:
                        # there was just a stray semicolon
                        # at the end, producing an
                        # emptystring
                        pass
                    else:
                        raise e
                else:
                    style[key] = value

        return style

    def parse_style(self, stylestr: str) -> Dict:
        # Originally written by @MrMallIronmaker
        # Copied from Manim.
        # https://github.com/ManimCommunity/manim/
        # blob/aa873e0f8e5c4f35c864568213b04b88af5b4769/manim/mobject/svg/style_utils.py#L143
        SVG_DEFAULT_ATTRIBUTES = self.SVG_DEFAULT_ATTRIBUTES
        manim_style = {}
        svg_style = self.str2styledict(stylestr)
        for key in SVG_DEFAULT_ATTRIBUTES:
            if key not in svg_style:
                svg_style[key] = SVG_DEFAULT_ATTRIBUTES[key]

        if "fill-opacity" in svg_style:
            manim_style["fill_opacity"] = float(svg_style["fill-opacity"])

        if "stroke-opacity" in svg_style:
            manim_style["stroke_opacity"] = float(svg_style["stroke-opacity"])

        # nones need to be handled specially
        if "fill" in svg_style:
            if svg_style["fill"] == "none":
                manim_style["fill_opacity"] = 0
            else:
                manim_style["fill_color"] = self.parse_color_string(svg_style["fill"])

        if "stroke" in svg_style:
            if svg_style["stroke"] == "none":
                # In order to not break animations.creation.Write,
                # we interpret no stroke as stroke-width of zero and
                # color the same as the fill color, if it exists.
                manim_style["stroke_width"] = 0
                if "fill_color" in manim_style:
                    manim_style["stroke_color"] = manim_style["fill_color"]
            else:
                manim_style["stroke_color"] = self.parse_color_string(
                    svg_style["stroke"]
                )

        return manim_style


class SVGStyleTester(SVGUtils):
    def __init__(self, gotSVG: Path, expectedSVG: Path) -> None:
        self.expected_file = expectedSVG
        self.got_file = gotSVG
        self.prepare()

    def get_handler(self, final_setter):
        def func(name, attrs):
            styles = []
            if name == "g":
                if "style" in attrs:
                    styles.append(self.parse_style(attrs["style"]))
            if styles:
                # need to think of better way than this.
                exec(f"self.{final_setter} += [styles]")

        return func

    def prepare(self):
        self.got_svg_style = []
        self.expected_svg_style = []
        got_parser = ParserCreate()
        got_parser.StartElementHandler = self.get_handler("got_svg_style")
        with open(self.got_file, "rb") as f:
            got_parser.ParseFile(f)

        expected_parser = ParserCreate()
        expected_parser.StartElementHandler = self.get_handler("expected_svg_style")
        with open(self.expected_file, "rb") as f:
            expected_parser.ParseFile(f)



cdef extern from "cairo.h":
    ctypedef struct cairo_surface_t:
        pass
    ctypedef struct cairo_t:
        pass
    ctypedef enum cairo_status_t:
        CAIRO_STATUS_SUCCESS
        CAIRO_STATUS_NO_MEMORY
    cairo_t* cairo_create(cairo_surface_t* target)
    void cairo_move_to(
        cairo_t* cr,
        double x,
        double y
    )
    void cairo_destroy(cairo_t* cr)
    void cairo_surface_destroy(cairo_surface_t* surface)

    cairo_status_t cairo_status(cairo_t *cr)
    const char* cairo_status_to_string(cairo_status_t status)
    const char* cairo_version_string()

cdef extern from "cairo-svg.h":
    cairo_surface_t* cairo_svg_surface_create(
        const char* filename,
        double width_in_points,
        double height_in_points
    )



# -*- coding: utf-8 -*-
import re

from .enums import Style, Weight


class PangoUtils:
    @staticmethod
    def str2style(string: str) -> Style:
        """Internally used function. Converts text to Pango Understandable Styles."""
        styles = {
            "NORMAL": Style.NORMAL,
            "ITALIC": Style.ITALIC,
            "OBLIQUE": Style.OBLIQUE,
        }
        try:
            return styles[string]
        except KeyError:
            raise AttributeError("There is no Style Called %s" % string)

    @staticmethod
    def str2weight(string: str) -> Weight:
        """Internally used function. Convert text to Pango Understandable Weight"""
        weights = {
            "NORMAL": Weight.NORMAL,
            "BOLD": Weight.BOLD,
            "THIN": Weight.THIN,
            "ULTRALIGHT": Weight.ULTRALIGHT,
            "LIGHT": Weight.LIGHT,
            "BOOK": Weight.BOOK,
            "MEDIUM": Weight.MEDIUM,
            "SEMIBOLD": Weight.SEMIBOLD,
            "ULTRABOLD": Weight.ULTRABOLD,
            "HEAVY": Weight.HEAVY,
            "ULTRAHEAVY": Weight.ULTRAHEAVY,
        }
        try:
            return weights[string]
        except KeyError:
            raise AttributeError("There is no Font Weight Called %s" % string)

    @staticmethod
    def remove_last_M(file_name: str) -> None:
        """Remove element from the SVG file in order to allow comparison."""
        with open(file_name, "r") as fpr:
            content = fpr.read()
        content = re.sub(r'Z M [^A-Za-z]*? "\/>', 'Z "/>', content)
        with open(file_name, "w") as fpw:
            fpw.write(content)



from pathlib import Path
from pango cimport *
import copy

cpdef bint fc_register_font(str font_path):
    """This function registers the font file using ``fontconfig`` so that
    it is available for use by Pango. On Linux it is aliased to
    :func:`register_font` and on Windows and macOS this would work only when
    using ``fontconfig`` backend.

    Parameters
    ==========
    font_path : :class:`str`
        Relative or absolute path to font file.
    Returns
    =======
    :class:`bool`
            True means it worked without any error.
            False means there was an unknown error
    Examples
    --------
    >>> register_font("/home/roboto.tff")
    True

    Raises
    ------
    AssertionError
        Font is missing.
    """
    a=Path(font_path)
    assert a.exists(), f"font doesn't exist at {a.absolute()}"
    font_path = str(a.absolute())
    font_path_bytes=font_path.encode('utf-8')
    cdef const unsigned char* fontPath = font_path_bytes
    fontAddStatus = FcConfigAppFontAddFile(FcConfigGetCurrent(), fontPath)
    if fontAddStatus:
        return True
    else:
        return False


cpdef bint fc_unregister_font(str font_path):
    """This function unregisters(removes) the font file using
    ``fontconfig``. It is mostly optional to call this.
    Mainly used in tests. On Linux it is aliased to
    :func:`unregister_font` and on Windows and macOS this
    would work only when using ``fontconfig`` backend.

    Parameters
    ==========

    font_path: :class:`str`
        For compatibility with the windows function.

    Returns
    =======
    :class:`bool`
            True means it worked without any error.
            False means there was an unknown error

    """
    FcConfigAppFontClear(NULL)
    return True


IF UNAME_SYSNAME == "Linux":
    register_font = fc_register_font
    unregister_font = fc_unregister_font


ELIF UNAME_SYSNAME == "Windows":
    cpdef bint register_font(str font_path):
        """This function registers the font file using native windows API
        so that it is available for use by Pango.

        Parameters
        ==========
        font_path : :class:`str`
            Relative or absolute path to font file.
        Returns
        =======
        :class:`bool`
                True means it worked without any error.
                False means there was an unknown error
        Examples
        --------
        >>> register_font("C:/home/roboto.tff")
        True
        Raises
        ------
        AssertionError
            Font is missing.
        """
        a=Path(font_path)
        assert a.exists(), f"font doesn't exist at {a.absolute()}"
        font_path = str(a.absolute())
        fontAddStatus = AddFontResourceExW(
            <bytes>font_path,
            FR_PRIVATE,
            0
        )
        if fontAddStatus > 0:
            return True
        else:
            return False


    cpdef bint unregister_font(str font_path):
        """This function unregisters(removes) the font file using
        native Windows API. It is mostly optional to call this.
        Mainly used in tests.
        Parameters
        ==========
        font_path : :class:`str`
            Relative or absolute path to font file.
        Returns
        =======
        :class:`bool`
                True means it worked without any error.
                False means there was an unknown error
        Raises
        ------
        AssertionError
            Font is missing.
        """
        a=Path(font_path)
        assert a.exists(), f"font doesn't exist at {a.absolute()}"
        font_path = str(a.absolute())
        return RemoveFontResourceExW(
            <bytes>font_path,
            FR_PRIVATE,
            0
        )


ELIF UNAME_SYSNAME == "Darwin":
    cpdef bint register_font(str font_path):
        """This function registers the font file using ``CoreText`` API so that
        it is available for use by Pango.
        Parameters
        ==========
        font_path : :class:`str`
            Relative or absolute path to font file.
        Returns
        =======
        :class:`bool`
                True means it worked without any error.
                False means there was an unknown error
        Examples
        --------
        >>> register_font("/home/roboto.tff")
        True
        Raises
        ------
        AssertionError
            Font is missing.
        """
        a = Path(font_path)
        assert a.exists(), f"font doesn't exist at {a.absolute()}"
        font_path_bytes_py = str(a.absolute().as_uri()).encode('utf-8')
        cdef unsigned char* font_path_bytes = <bytes>font_path_bytes_py
        b = len(a.absolute().as_uri())
        cdef CFURLRef cf_url = CFURLCreateWithBytes(NULL, font_path_bytes, b, 0x08000100, NULL)
        return CTFontManagerRegisterFontsForURL(
            cf_url,
            kCTFontManagerScopeProcess,
            NULL
        )


    cpdef bint unregister_font(str font_path):
        """This function unregisters(removes) the font file using
        native ``CoreText`` API. It is mostly optional to call this.
        Mainly used in tests.
        Parameters
        ==========
        font_path : :class:`str`
            Relative or absolute path to font file.
        Returns
        =======
        :class:`bool`
                True means it worked without any error.
                False means there was an unknown error
        Raises
        ------
        AssertionError
            Font is missing.
        """
        a = Path(font_path)
        assert a.exists(), f"font doesn't exist at {a.absolute()}"
        font_path_bytes_py = str(a.absolute().as_uri()).encode('utf-8')
        cdef unsigned char* font_path_bytes = <bytes>font_path_bytes_py
        b = len(a.absolute().as_uri())
        cdef CFURLRef cf_url = CFURLCreateWithBytes(NULL, font_path_bytes, b, 0x08000100, NULL)
        return CTFontManagerUnregisterFontsForURL(
            cf_url,
            kCTFontManagerScopeProcess,
            NULL
        )


cpdef list list_fonts():
    """Lists the fonts available to Pango.
    This is usually same as system fonts but it also
    includes the fonts added through :func:`register_font`.

    Returns
    -------

    :class:`list` :
        List of fonts sorted alphabetically.
    """
    cdef PangoFontMap* fontmap=pango_cairo_font_map_new()
    if fontmap == NULL:
        raise MemoryError("Pango.FontMap can't be created.")
    cdef int n_families=0
    cdef PangoFontFamily** families=NULL
    pango_font_map_list_families(
        fontmap,
        &families,
        &n_families
    )
    if families is NULL or n_families==0:
        raise MemoryError("Pango returned unexpected length on families.")
    family_list=[]
    for i in range(n_families):
        name = copy.deepcopy(pango_font_family_get_name(families[i]).decode('utf-8'))
        # according to pango's docs, the `char *` returned from
        # `pango_font_family_get_name`is owned by pango, and python
        # shouldn't interfere with it. So, rather we are making a
        # deepcopy so that we don't worry about it.
        family_list.append(name)
    g_free(families)
    g_object_unref(fontmap)
    family_list.sort()
    return family_list



# -*- coding: utf-8 -*-
import os
import subprocess
import sys
import textwrap
from contextlib import contextmanager
from pathlib import Path
from shutil import copyfile

import pytest

import manimpango

from . import FONT_DIR
from ._manim import MarkupText


@contextmanager
def register_font(font_file: Path):
    font_file = os.fspath(font_file)
    init = manimpango.list_fonts()
    assert manimpango.register_font(font_file), "Invalid Font possibly."
    final = manimpango.list_fonts()
    yield list(set(final) - set(init))[0]
    assert manimpango.unregister_font(font_file), "Can't unregister Font"


def font_list():
    _t = [
        (FONT_DIR / "AdobeVFPrototype.ttf").absolute(),
        (FONT_DIR / "BungeeColor-Regular_colr_Windows.ttf").absolute(),
        (FONT_DIR / "MaShanZheng-Regular.ttf").absolute(),
    ]
    _d = {}
    for i in _t:
        with register_font(i) as f:
            _d[i] = f
    return _d


font_lists_dict = font_list()


def test_unicode_font_name(tmpdir):
    final_font = str(Path(tmpdir, "庞门正.ttf").absolute())
    copyfile(FONT_DIR / "AdobeVFPrototype.ttf", final_font)
    assert manimpango.register_font(final_font)
    assert manimpango.unregister_font(final_font)


@pytest.mark.parametrize("font_name", font_lists_dict)
def test_register_and_unregister_font(font_name):
    intial = manimpango.list_fonts()
    assert manimpango.register_font(str(font_name)), "Invalid Font possibly."
    final = manimpango.list_fonts()
    assert intial != final
    assert manimpango.unregister_font(os.fspath(font_name)), "Can't unregister font."
    assert intial == manimpango.list_fonts()


@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="no warning are raised for linux"
)
@pytest.mark.parametrize("font_file,font_name", font_lists_dict.items())
def test_warning(font_file, font_name):
    # this tests need to be run in a separate subprocess because
    # of fontmap cache in Pango.
    command = textwrap.dedent(
        f"""\
            from tests import set_dll_search_path
            set_dll_search_path() # this is needed on Windows to run
            import manimpango
            import os
            from tests._manim import Text
            font_file = r'{os.fspath(font_file)}'
            font_name = r'{font_name}'
            intial = manimpango.list_fonts()
            manimpango.register_font(os.fspath(font_file))
            final = manimpango.list_fonts()
            assert intial != final
            Text("Testing", font=font_name)
            manimpango.unregister_font(os.fspath(font_file))
        """
    )
    a = subprocess.run(
        [sys.executable, "-c", command],
        check=True,
        stderr=subprocess.PIPE,
        cwd=Path(__file__).parent.parent,
    )
    captured = a.stderr.decode()
    assert "Pango-WARNING **" not in captured, "Looks like Pango raised a warning?"


@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="unsupported api for linux"
)
@pytest.mark.parametrize("font_name", font_lists_dict)
@pytest.mark.skipif(sys.platform.startswith("darwin"), reason="always returns true")
def test_fail_just_unregister(font_name):
    assert not manimpango.unregister_font(
        str(font_name)
    ), "Failed to unregister the font"


@pytest.mark.skipif(
    sys.platform.startswith("win32"), reason="unsupported api for win32"
)
@pytest.mark.skipif(sys.platform.startswith("darwin"), reason="unsupported api for mac")
def test_unregister_not_fail_linux():
    assert manimpango.unregister_font("random")


@pytest.mark.skipif(
    sys.platform.startswith("linux"), reason="unsupported api for linux"
)
def test_adding_dummy_font(tmpdir):
    dummy = tmpdir / "font.ttf"
    with open(dummy, "wb") as f:
        f.write(b"dummy")
    assert not manimpango.register_font(str(dummy)), "Registered a dummy font?"
    assert not manimpango.fc_register_font(str(dummy)), "Registered a dummy font?"


def test_simple_fonts_render(tmpdir):
    filename = str(Path(tmpdir) / "hello.svg")
    MarkupText("Hello World", filename=filename)
    assert Path(filename).exists()


@pytest.mark.skipif(
    not sys.platform.startswith("linux"), reason="unsupported api other than linux"
)
def test_both_fc_and_register_font_are_same():
    assert manimpango.fc_register_font == manimpango.register_font
    assert manimpango.fc_unregister_font == manimpango.unregister_font


@pytest.mark.parametrize("font_file", font_lists_dict)
def test_fc_font_register(setup_fontconfig, font_file):
    intial = manimpango.list_fonts()
    assert manimpango.fc_register_font(str(font_file)), "Invalid Font possibly."
    final = manimpango.list_fonts()
    assert intial != final


def test_fc_font_unregister(setup_fontconfig):
    # it will remove everything
    intial = manimpango.list_fonts()
    manimpango.fc_unregister_font("clear")
    final = manimpango.list_fonts()
    assert intial != final



cdef extern from "glib.h":
    ctypedef void* gpointer
    ctypedef int gint
    ctypedef unsigned int guint
    ctypedef gint gboolean
    ctypedef unsigned short guint16
    ctypedef char gchar
    ctypedef struct GError:
        gint code
        gchar *message
    void g_error_free (GError *error)
    void g_object_unref(gpointer object)
    void g_free(gpointer mem)




cdef extern from "fontconfig/fontconfig.h":
    ctypedef int FcBool
    ctypedef struct FcConfig:
        pass
    FcBool FcConfigAppFontAddFile(
        FcConfig* config,
        const unsigned char* file_name
    )
    FcConfig* FcConfigGetCurrent()
    void FcConfigAppFontClear(void *)

# Windows and macOS specific API's
IF UNAME_SYSNAME == "Windows":
    cdef extern from "windows.h":
        ctypedef Py_UNICODE WCHAR
        ctypedef const WCHAR* LPCWSTR
        ctypedef enum DWORD:
            FR_PRIVATE
        int AddFontResourceExW(
            LPCWSTR name,
            DWORD fl,
            unsigned int res
        )
        bint RemoveFontResourceExW(
            LPCWSTR name,
            DWORD  fl,
            unsigned int pdv
        )
ELIF UNAME_SYSNAME == "Darwin":
    cdef extern from "Carbon/Carbon.h":
        ctypedef struct CFURLRef:
            pass
        ctypedef enum CTFontManagerScope:
            kCTFontManagerScopeProcess
        ctypedef unsigned int UInt8
        ctypedef long CFIndex
        ctypedef unsigned int UInt32
        ctypedef UInt32 CFStringEncoding
        CFURLRef CFURLCreateWithBytes(
            void*,
            unsigned char *URLBytes,
            CFIndex length,
            CFStringEncoding encoding,
            void*
        )
        bint CTFontManagerRegisterFontsForURL(
            CFURLRef fontURL,
            CTFontManagerScope scope,
            void* error
        )
        bint CTFontManagerUnregisterFontsForURL(
            CFURLRef fontURL,
            CTFontManagerScope scope,
            void* error
        )



# -*- coding: utf-8 -*-

import argparse
import os
import shlex
import subprocess
import sys
import warnings
from pathlib import Path
from shlex import quote
from subprocess import PIPE, Popen, check_call

from setuptools import Extension, setup

try:
    from Cython.Build import cythonize

    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

coverage = False
if sys.argv[-1] == "--coverage":
    coverage = True
    sys.argv.pop()


def get_version():
    version_file = "manimpango/_version.py"
    with open(version_file) as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


NAME = "ManimPango"
MANIMPANGO_VERSION = get_version()
MINIMUM_PANGO_VERSION = "1.30.0"
DEBUG = False

if sys.platform == "win32" and sys.version_info >= (3, 10):
    import atexit

    atexit.register(
        lambda: warnings.warn(
            f"ManimPango {MANIMPANGO_VERSION} does not support Python "
            f"{sys.version_info.major}.{sys.version_info.minor} and does not provide "
            "prebuilt Windows binaries. We do not recommend building from source on "
            "Windows.",
            RuntimeWarning,
        )
    )


class RequiredDependencyException(Exception):
    pass


NEEDED_LIBS = [
    "pangocairo-1.0",
    "pango-1.0",
    "gobject-2.0",
    "glib-2.0",
    "intl",
    "harfbuzz",
    "cairo",
    "fontconfig",
]


class PKG_CONFIG:
    """Preform operations with pkg-config.
    Parameters
    ==========
    libname : :class:`str`
        The library name to query.
    """

    def __init__(self, libname: str) -> None:
        self.name = libname
        self.pkg_config = os.environ.get("PKG_CONFIG", "pkg-config")
        self.setup_argparse()
        self.check_pkgconfig = self.check_pkgconfig()

    def check_pkgconfig(self):
        """Check whether pkg-config works.
        If it doesn't work raise a warning.
        """
        try:
            check_call([self.pkg_config, "--version"], stdout=subprocess.DEVNULL)
            return True
        except Exception:
            warnings.warn(
                f"{self.pkg_config} doesn't exists or doesn't seem to work\n"
                "We assume that you give the compiler flags using\n"
                "corresponding environment variables.",
                RuntimeWarning,
            )
            return False

    def parse_cflags(self, raw_cflags):
        """Parse the flags from pkg-config using argparse."""
        raw_args = shlex.split(raw_cflags.strip())
        args, unknown = self._cflag_parser.parse_known_args(raw_args)
        config = {k: v or [] for k, v in args.__dict__.items()}
        for i, x in enumerate(config["define_macros"]):
            parts = x.split("=", 1)
            value = x[1] or None if len(x) == 2 else None
            config["define_macros"][i] = (parts[0], value)
        return config, " ".join(quote(x) for x in unknown)

    def setup_argparse(self):
        _cflag_parser = argparse.ArgumentParser(add_help=False)
        _cflag_parser.add_argument("-I", dest="include_dirs", action="append")
        _cflag_parser.add_argument("-L", dest="library_dirs", action="append")
        _cflag_parser.add_argument("-l", dest="libraries", action="append")
        _cflag_parser.add_argument("-D", dest="define_macros", action="append")
        _cflag_parser.add_argument("-R", dest="runtime_library_dirs", action="append")
        self._cflag_parser = _cflag_parser

    def check_min_version(self, version):
        """Check whether the library of that version exists."""
        command = [
            self.pkg_config,
            "--print-errors",
            "--atleast-version",
            version,
            self.name,
        ]
        try:
            check_call(command, stdout=subprocess.DEVNULL)
            return True
        except Exception:
            raise RequiredDependencyException(f"{self.name} >= {version} is required")

    @property
    def libs(self):
        """Get distutils-compatible extension extras for the given library.
        This requires ``pkg-config``.
        """
        if self.check_pkgconfig:
            name = self.name
            command = self.pkg_config
            try:
                proc = Popen(
                    [command, "--libs", name],
                    stdout=PIPE,
                    stderr=PIPE,
                )
            except Exception:
                pass
            raw_libs, _ = proc.communicate()
            known_libs, unknown_libs = self.parse_cflags(raw_libs.decode("utf8"))
            if unknown_libs:
                known_libs["extra_link_args"] = unknown_libs.split()
            return known_libs
        return dict()

    @property
    def cflags(self):
        if self.check_pkgconfig:
            name = self.name
            command = self.pkg_config
            try:
                proc = Popen(
                    [command, "--cflags", name],
                    stdout=PIPE,
                    stderr=PIPE,
                )
            except Exception:
                pass
            raw_cflags, _ = proc.communicate()
            known_cflags, unknown_cflags = self.parse_cflags(raw_cflags.decode("utf8"))
            if unknown_cflags:
                known_cflags["extra_compile_args"] = unknown_cflags.split()
            return known_cflags
        return dict()

    @property
    def setuptools_args(self):
        return update_dict(self.libs, self.cflags)


def update_dict(dict1: dict, dict2: dict):
    for key in dict1:
        if key in dict2:
            dict2[key] = dict1[key] + dict2[key]
        else:
            dict2[key] = dict1[key]
    return dict2


ext = ".pyx" if USE_CYTHON else ".c"
base_file = Path(__file__).parent / "manimpango"
_pkg_config_pangocairo = PKG_CONFIG("pangocairo")
_pkg_config_fontconfig = PKG_CONFIG("pangofc")
if _pkg_config_pangocairo.check_pkgconfig:
    _pkg_config_pangocairo.check_min_version(MINIMUM_PANGO_VERSION)
    returns = update_dict(
        _pkg_config_pangocairo.setuptools_args, _pkg_config_fontconfig.setuptools_args
    )
else:
    returns = {}
    returns["libraries"] = NEEDED_LIBS
if sys.platform == "win32":
    returns["libraries"] += ["Gdi32"]
    if hasattr(returns, "define_macros"):
        returns["define_macros"] += [("UNICODE", 1)]
    else:
        returns["define_macros"] = [("UNICODE", 1)]
if coverage:
    returns["define_macros"] += [("CYTHON_TRACE", 1)]
    returns["define_macros"] += [("CYTHON_TRACE_NOGIL", 1)]

ext_modules = [
    Extension(
        "manimpango.cmanimpango",
        [str(base_file / ("cmanimpango" + ext))],
        **returns,
    ),
    Extension(
        "manimpango.enums",
        [str(base_file / ("enums" + ext))],
        **returns,
    ),
    Extension(
        "manimpango.register_font",
        [str(base_file / ("register_font" + ext))],
        **returns,
    ),
]
if USE_CYTHON:
    ext_modules = cythonize(
        ext_modules,
        language_level=3,
        include_path=["manimpango"],
        gdb_debug=DEBUG,
        compiler_directives={"linetrace": coverage},
    )
with open("README.md") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=MANIMPANGO_VERSION,
    author="Naveen M K",
    author_email="naveen@manim.community",
    maintainer="The Manim Community Developers",
    url="https://manimpango.manim.community/",
    description="Bindings for Pango for using with Manim.",
    long_description=long_description,
    zip_safe=False,
    long_description_content_type="text/markdown",
    packages=["manimpango"],
    python_requires=">=3.7",
    platforms=["Linux", "macOS", "Windows"],
    keywords=["cython", "pango", "cairo", "manim"],
    license="GPL-3.0-or-later",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Cython",
    ],
    project_urls={
        "Documentation": "https://manimpango.manim.community/",
        "Source": "https://github.com/ManimCommunity/manimpango",
        "Release notes": "https://github.com/ManimCommunity/ManimPango/releases/",
    },
    ext_modules=ext_modules,
    package_data={
        "manimpango": ["*.pxd", "*.pyx"],
    },
)



# -*- coding: utf-8 -*-
from pathlib import Path

import pytest

import manimpango

from . import CASES_DIR
from ._manim import MarkupText
from .svg_tester import SVGStyleTester

ipsum_text = (
    "<b>Lorem ipsum dolor</b> sit amet, <i>consectetur</i> adipiscing elit,"
    "sed do eiusmod tempor incididunt ut labore et dolore"
    "magna aliqua. Ut enim <b>ad</b> minim veniam, quis nostrud"
    "exercitation ullamco laboris nisi ut aliquip"
    "ex ea commodo consequat. Duis aute irure dolor"
    "in reprehenderit in voluptate velit esse cillum"
    "dolore eu fugiat nulla pariatur. Excepteur sint"
    "occaecat cupidatat non proident, sunt in culpa qui"
    "officia deserunt mollit anim id est laborum."
)


@pytest.mark.parametrize("text", ["foo", "<b>bar</b>", "வணக்கம்"])
def test_good_markup(text):

    assert not manimpango.MarkupUtils.validate(
        text,
    ), f"{text} should not fail validation"


@pytest.mark.parametrize("text", ["<b>foo", "<xyz>foo</xyz>"])
def test_bad_markup(text):
    assert manimpango.MarkupUtils.validate(
        text
    ), f"{text} should fail validation (unbalanced tags)"


@pytest.mark.parametrize(
    "text,error",
    [
        (
            "<b>foo",
            "Error on line 1 char 23: Element “markup” was closed, "
            "but the currently open element is “b”",
        ),
        (
            "<xyz>foo</xyz>",
            "Unknown tag 'xyz' on line 1 char 14",
        ),
    ],
)
def test_bad_markup_error_message(text, error):
    assert manimpango.MarkupUtils.validate(text) == error


def test_markup_text(tmpdir):
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    MarkupText(
        '<span underline="error"><b><i>Hello Manim</i></b></span>', filename=str(loc)
    )
    assert loc.exists()


def test_markup_justify(tmpdir):
    # don't know how to verify this correctly
    # it varies upon diffent system so, we are
    # just check whether it runs
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    MarkupText(ipsum_text, justify=True, filename=str(loc))
    assert loc.exists()


def test_markup_indent(tmpdir):
    # don't know how to verify this correctly
    # it varies upon diffent system so, we are
    # just check whether it runs
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    MarkupText(ipsum_text, indent=10, filename=str(loc))
    assert loc.exists()


def test_markup_alignment(tmpdir):
    # don't know how to verify this correctly
    # it varies upon diffent system so, we are
    # just check whether it runs
    loc = Path(tmpdir, "test.svg")
    assert not loc.exists()
    MarkupText(
        ipsum_text,
        alignment=manimpango.Alignment.CENTER,
        filename=str(loc),
    )
    assert loc.exists()


def test_markup_style(tmpdir):
    test_case = CASES_DIR / "hello_blue_world_green.svg"
    expected = tmpdir / "expected.svg"
    text = "<span foreground='BLUE'>Hello</span>\n<span foreground='GREEN'>World</span>"
    MarkupText(
        text,
        filename=str(expected),
    )
    s = SVGStyleTester(gotSVG=expected, expectedSVG=test_case)
    assert len(s.got_svg_style) == len(s.expected_svg_style)
    assert s.got_svg_style == s.expected_svg_style


def test_wrap_text(tmpdir):
    tmpdir = Path(tmpdir)
    wrapped = tmpdir / "wrap.svg"
    nowrap = tmpdir / "nowarap.svg"

    MarkupText(ipsum_text, wrap_text=False, filename=str(nowrap))
    MarkupText(ipsum_text, filename=str(wrapped))

    assert wrapped.read_text() != nowrap.read_text()



# -*- coding: utf-8 -*-
def test_pango_version():
    import manimpango

    v = manimpango.pango_version()
    assert type(v) == str


def test_cairo_version():
    import manimpango

    v = manimpango.cairo_version()
    assert type(v) == str



# -*- coding: utf-8 -*-
import sys

import pytest

import manimpango

from .test_fonts import font_lists_dict as font_lists


def test_whether_list():
    a = manimpango.list_fonts()
    assert type(a) is list
    assert len(a) > 0


@pytest.mark.skipif(
    sys.platform.startswith("linux") or sys.platform.startswith("darwin"),
    reason="names don't match for some reason.",
)
@pytest.mark.parametrize("font_file", font_lists)
def test_resgister_font_with_list(font_file):
    manimpango.register_font(str(font_file))
    a = manimpango.list_fonts()
    assert font_lists[font_file] in a
    manimpango.unregister_font(str(font_file))



# -*- coding: utf-8 -*-
"""This file contains helpers for the tests copied and modified
from Manim.
"""
import copy
import os
import re
from pathlib import Path
from typing import Sequence

from manimpango import Alignment, MarkupUtils, TextSetting, text2svg


class MarkupText:
    def __init__(
        self,
        text: str,
        *,
        size: int = 1,
        line_spacing: int = None,
        font: str = None,
        slant: str = "NORMAL",
        weight: str = "NORMAL",
        tab_width: int = 4,
        disable_ligatures: bool = False,
        justify: bool = None,
        indent: float = None,
        alignment: Alignment = None,
        # for the tests
        filename: str = "test.svg",
        wrap_text: bool = True,
        **kwargs,
    ):
        self.text = text
        self.size = size
        self.line_spacing = line_spacing
        self.font = font
        self.slant = slant
        self.weight = weight
        self.tab_width = tab_width
        self.filename = filename
        self.original_text = text
        self.disable_ligatures = disable_ligatures

        self.justify = justify
        self.indent = indent
        self.alignment = alignment
        self.wrap_text = wrap_text
        if MarkupUtils.validate(self.text):
            raise ValueError(
                f"Pango cannot parse your markup in {self.text}. "
                "Please check for typos, unmatched tags or unescaped "
                "special chars like < and &."
            )
        self.text2svg()

    def text2svg(self):
        """Convert the text to SVG using Pango."""
        size = self.size * 10
        dir_name = Path(self.filename).parent
        disable_liga = self.disable_ligatures
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        file_name = self.filename
        if self.wrap_text:
            return MarkupUtils.text2svg(
                f"{self.text}",
                self.font,
                self.slant,
                self.weight,
                size,
                True,  # stray positional argument
                disable_liga,
                file_name,
                20,
                20,
                600,  # width
                400,  # height
                justify=self.justify,
                indent=self.indent,
                line_spacing=self.line_spacing,
                alignment=self.alignment,
            )
        else:
            return MarkupUtils.text2svg(
                f"{self.text}",
                self.font,
                self.slant,
                self.weight,
                size,
                True,  # stray positional argument
                disable_liga,
                file_name,
                20,
                20,
                600,  # width
                400,  # height
                justify=self.justify,
                indent=self.indent,
                line_spacing=self.line_spacing,
                alignment=self.alignment,
                pango_width=-1,
            )
            # -1 for no wrapping
            # default is full width and then wrap.

    def __repr__(self):
        return f"MarkupText({repr(self.original_text)})"


class Text:
    def __init__(
        self,
        text: str,
        fill_opacity: float = 1.0,
        stroke_width: int = 0,
        size: int = 1,
        line_spacing: int = -1,
        font: str = "",
        slant: str = "NORMAL",
        weight: str = "NORMAL",
        gradient: tuple = None,
        tab_width: int = 4,
        disable_ligatures: bool = False,
        filename: str = "text.svg",
        color: str = "white",
        **kwargs,
    ) -> None:
        self.color = color
        self.size = size
        self.filename = filename
        self.line_spacing = line_spacing
        self.font = font
        self.slant = slant
        self.weight = weight
        self.gradient = gradient
        self.tab_width = tab_width
        self.original_text = text
        self.disable_ligatures = disable_ligatures
        text_without_tabs = text
        self.t2f = self.t2s = self.t2w = self.t2c = {}
        if text.find("\t") != -1:
            text_without_tabs = text.replace("\t", " " * self.tab_width)
        self.text = text_without_tabs
        if self.line_spacing == -1:
            self.line_spacing = self.size + self.size * 0.3
        else:
            self.line_spacing = self.size + self.size * self.line_spacing
        self.text2svg()

    def _merge_settings(
        self, left_setting: TextSetting, right_setting: TextSetting, args: Sequence[str]
    ) -> TextSetting:
        contained = right_setting.end < left_setting.end
        new_setting = copy.copy(left_setting) if contained else copy.copy(right_setting)

        new_setting.start = right_setting.end if contained else left_setting.end
        left_setting.end = right_setting.start
        if not contained:
            right_setting.end = new_setting.start

        for arg in args:
            left = getattr(left_setting, arg)
            right = getattr(right_setting, arg)
            default = getattr(self, arg)
            if left != default and getattr(right_setting, arg) != default:
                raise ValueError(
                    "Ambiguous style for text"
                    f"'{self.text[right_setting.start:right_setting.end]}':"
                    f"'{arg}' cannot be both '{left}' and '{right}'."
                )
            setattr(right_setting, arg, left if left != default else right)
        return new_setting

    def text2settings(self):
        """Internally used function. Converts the texts and styles
        to a setting for parsing."""
        settings = []
        t2xs = [
            (self.t2f, "font"),
            (self.t2s, "slant"),
            (self.t2w, "weight"),
            (self.t2c, "color"),
        ]
        t2xwords = {
            *self.t2f.keys(),
            *self.t2s.keys(),
            *self.t2w.keys(),
            *self.t2c.keys(),
        }
        for word in t2xwords:
            setting_args = {
                arg: t2x[word] if word in t2x else getattr(self, arg)
                for t2x, arg in t2xs
            }

            for start, end in self.find_indexes(word, self.text):
                settings.append(TextSetting(start, end, **setting_args))

        # Handle overlaps
        setting_args = {arg: getattr(self, arg) for _, arg in t2xs}

        settings.sort(key=lambda setting: setting.start)
        index = 0
        for setting in settings:
            index += 1
            if index == len(settings):
                break

            next_setting = settings[index]
            if setting.end > next_setting.start:
                new_setting = self._merge_settings(setting, next_setting, setting_args)
                new_index = index
                while (
                    new_index < len(settings)
                    and settings[new_index].start < new_setting.start
                ):
                    new_index += 1
                settings.insert(new_index, new_setting)

        # Set all text settings (default font, slant, weight)
        temp_settings = settings.copy()
        start = 0
        for setting in settings:
            if setting.start != start:
                temp_settings.append(TextSetting(start, setting.start, **setting_args))
            start = setting.end
        if start != len(self.text):
            temp_settings.append(TextSetting(start, len(self.text), **setting_args))
        settings = sorted(temp_settings, key=lambda setting: setting.start)

        if re.search(r"\n", self.text):
            line_num = 0
            for start, end in self.find_indexes("\n", self.text):
                for setting in settings:
                    if setting.line_num == -1:
                        setting.line_num = line_num
                    if start < setting.end:
                        line_num += 1
                        new_setting = copy.copy(setting)
                        setting.end = end
                        new_setting.start = end
                        new_setting.line_num = line_num
                        settings.append(new_setting)
                        settings.sort(key=lambda setting: setting.start)
                        break
        for setting in settings:
            if setting.line_num == -1:
                setting.line_num = 0
        return settings

    def text2svg(self):
        """Internally used function.
        Convert the text to SVG using Pango
        """
        size = self.size * 10
        line_spacing = self.line_spacing * 10
        dir_name = Path(self.filename).parent
        disable_liga = self.disable_ligatures
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        file_name = self.filename
        settings = self.text2settings()
        width = 600
        height = 400

        return text2svg(
            settings,
            size,
            line_spacing,
            disable_liga,
            file_name,
            30,
            30,
            width,
            height,
            self.text,
        )
